; Acl2 Version 1.5

; Copyright (C) 1989-94 Computational Logic, Inc. (CLI).  All rights reserved.

; Use of this software constitutes agreement with the terms of the Acl2 GENERAL
; PUBLIC SOFTWARE LICENSE.  The license agreement is displayed at the beginning
; of the Acl2 file "acl2.lisp".

; Acl2 (A Computational Logic for an Applicative Core Language) was written at
; Computational Logic, Inc. by

;   Robert S. Boyer,
;   Matt Kauffmann, and
;   J Strother Moore

; Acknowledgements: Many CLI employees have contributed to the design of
; certain features.  We especially thank John Cowles for his contributions.

; Sponsorship: This work was supported in part at Computational Logic, Inc., by
; ??? contract numbers???  The views and conclusions contained in this document
; are those of the authors and should not be interpreted as representing the
; official policies, either expressed or implied, of Computational Logic, Inc.,
; ???, or the U.S. Government.

#|
------------------------------------------------------------------------------
                     Acl2 GENERAL PUBLIC SOFTWARE LICENSE
                           Computational Logic, Inc.
                          1717 West Sixth, Suite 290
                           Austin, Texas 78703-4776

Please read this license carefully before using the Acl2 Software.  By using
the Acl2 Software, you are agreeing to be bound by the terms of this license.
If you do not agree to the terms of this license, promptly return the Acl2
Software to the place where you obtained it.

The Acl2 Software was developed by Computational Logic, Inc.(CLI).  You own the
disk or other medium on which the Acl2 Software is recorded, but CLI retains
title to the Acl2 Software.  The purposes of this license are to identify the
Acl2 Software and to make the Acl2 Software, including its source code, freely
available.  This license allows you to use, copy, distribute and modify the
Acl2 Software, on the condition that you comply with all the Copying Policies
set out below.

COPYING POLICIES

1. You may copy and distribute verbatim copies of the Acl2 Software as you
receive it, in any medium, including embedding it verbatim in derivative works,
provided that you a) conspicuously and appropriately publish on each copy a
valid copyright notice "Copyright (C) 1989-94 by Computational Logic Inc.  All
Rights Reserved.", b) keep intact on all files the notices that refer to this
License Agreement and to the absence of any warranty, and c) give all
recipients of the Acl2 Software a copy of this License Agreement along with the
program.

2.  You may modify your copy or copies of the Acl2 Software or any portion of
it, and copy and distribute such modifications provided you tell recipients
that what they have is a modification by your organization of the CLI version
of the Acl2 Software.

3. You may incorporate parts of the Acl2 Software into other programs provided
that you

    a) acknowledge Computational Logic Inc. in the program documentation, and

    b) send a copy of a user's manual for the program to

       Software-Request                 or     Software-Request@CLI.COM
       Computational Logic Inc.
       1717 West Sixth, Suite 290
       Austin, TX 78703-4776

CLI also requests, but does not require, that any improvements or extensions to
the Acl2 Software be returned to one of these addresses, so that they may be
shared with other Acl2 users.  The Acl2 Software, including its source, can be
obtained by contacting one of the above addresses.

NO WARRANTY

BECAUSE THE Acl2 SOFTWARE IS LICENSED FREE OF CHARGE, WE PROVIDE ABSOLUTELY NO
WARRANTY.  THE SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE
RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE
Acl2 SOFTWARE PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING,
REPAIR OR CORRECTION.

IN NO EVENT WILL COMPUTATIONAL LOGIC INC. BE LIABLE TO YOU FOR ANY DAMAGES, ANY
LOST PROFITS, LOST MONIES, OR OTHER SPECIAL, INCIDENTAL OR CONSEQUENTIAL
DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE Acl2 SOFTWARE (INCLUDING
BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED INACCURATE OR LOSSES
SUSTAINED BY THIRD PARTIES), EVEN IF YOU HAVE ADVISED US OF THE POSSIBILITY OF
SUCH DAMAGES, OR FOR ANY CLAIM BY ANY OTHER PARTY.
------------------------------------------------------------------------------
|#

; This file, acl2.lisp, (a) builds the packages for the Acl2 system, (b)
; defines the functions for loading and compiling Acl2 and for running code
; verified using Acl2, and (c) makes a couple of checks concerning minor,
; non-CLTL, assumptions that we make about Common Lisps in which Acl2 is to
; run.

;                               FILES


; This is the file acl2.lisp, the first source file for Acl2.  The
; names of the other Acl2 source files are listed under *acl2-files*
; in the file axioms.lisp.

; All of Acl2 is written in Common Lisp and we expect that Acl2 will
; run in any Common Lisp except for those Common Lisps which fail the
; tests we execute upon loading this file, acl2.lisp.  With the
; exception of this file, the file interface-raw.lisp, and those
; constructs after the #-acl2-logic passim only (especially in the
; file axioms.lisp), Acl2 is written in the applicative core language
; for which Acl2 is a verification system.


;                            CLTL1/CLTL2

; For the most part, we program in a subset both of CLTL1 and CLTL2.
; However, there are a few places, notably the name of the main
; package for Lisp symbols, where this is impossible.  So we use the
; read conditional #+CLTL2.  If one wishes to run Acl2 under CLTL2,
; execute the following form before commencing compiling or loading:

;        (push :CLTL2 *features*)

; The theory will change as a result of this ``push'' in one important
; way:  the package-name of the main Lisp symbols will become
; "COMMON-LISP" instead of "LISP".

; We use IN-PACKAGE in a way that is consistent with both CLTL1 and
; CLTL2, namely as a macro (i.e., whose argument is not evaluated) for
; switching *package* to an existing package, value ignored.

(in-package #-CLTL2 "USER" #+CLTL2 "COMMON-LISP-USER")


;                              PACKAGES

; We never intend that this file should be compiled, and hence we do
; not need to obey the CLTL1 strictures about putting all IN-PACKAGE
; forms near the front.

(eval-when (compile)
           (error "The file acl2.lisp should never be compiled."))

(dolist
 (p (list-all-packages))
 (cond ((equal 3 (string< "ACL" (package-name p)))
        (format t
                "~%~%Warning:  There is already a package with the ~
                 name ~a. ~%The Acl2 implementation depends upon ~
                 having complete knowledge of ~%and control over any ~
                 packge whose name begins with the ~%three letters ``ACL'', ~
                 so Acl2 may not work in this Lisp." (package-name p))
        (cond ((package-use-list p)
               (format t "~%~%Warning:  The package with name ~a ~
                   USES the packages in the list ~a.  Acl2 will not work ~
                   in state of affairs."
                       (package-name p) (package-use-list p)))))))

(or (find-package "ACL2")
    (make-package "ACL2" :use nil))

(defmacro acl2::defconst (&rest acl2::x)
  (cons 'defparameter acl2::x))

; We now get our imports for package ACL2, putting them into the
; variable acl2::*common-lisp-symbols-from-main-lisp-package*.

(with-open-file
 (fl (open "axioms-akcl.lisp" :direction :input))

;  Get into the main lisp package in order to read in
;  *common-lisp-symbols-from-main-lisp-package*.

 (let ((*package* (find-package #-CLTL2 "LISP" #+CLTL2 "COMMON-LISP")))

;  Skip the in-package

   (read fl)

;  Do the defconst for *common-lisp-symbols-from-main-lisp-package*.

   (eval (read fl))
   (import acl2::*common-lisp-symbols-from-main-lisp-package* "ACL2")

;  Do the defconst for *acl2-version*.

   (eval (read fl))

;  Do the defconst for *acl2-files*.

   (eval (read fl))))

(in-package "ACL2")

(defparameter *global-package-prefix* "ACL2_GLOBAL_")

(defun make-global-package (x)
  (let* ((n (concatenate 'string *global-package-prefix* x))
         (p (find-package n)))
    (cond (p
           (do-symbols (sym p)
                       (makunbound sym)))
          (t (make-package n :use nil)))))

(or (find-package "ACL2-INPUT-CHANNEL")
    (make-package "ACL2-INPUT-CHANNEL" :use nil))

(or (find-package "ACL2-OUTPUT-CHANNEL")
    (make-package "ACL2-OUTPUT-CHANNEL" :use nil))


; Here are the initial global packages.

(make-global-package "LISP")
(make-global-package "KEYWORD")
(make-global-package "ACL2")
(make-global-package "ACL2-INPUT-CHANNEL")
(make-global-package "ACL2-OUTPUT-CHANNEL")


; The ACL2_INVISIBLE packages is used for a few odd objects that are
; to be hidden from the Acl2 user.

(let ((inv "ACL2_INVISIBLE"))
  (or (find-package inv)
      (make-package inv :use nil)))

; Common Lisp does not require that the symbol-package of the basic
; Lisp symbols be the basic LISP or COMMON-LISP package, but merely
; that those symbols be exported from that package.

(defparameter acl2::*initial-lisp-symbol-mark*
  'acl2_invisible::initial-lisp-symbol-mark)

(let ((p (find-package #-CLTL2 "LISP" #+CLTL2 "COMMON-LISP")))
  (dolist (sym *common-lisp-symbols-from-main-lisp-package*)
          (setf (get sym *initial-lisp-symbol-mark*)
                #-CLTL2 "LISP"
                #+CLTL2 "COMMON-LISP")
          (multiple-value-bind
           (s ans) (find-symbol (symbol-name sym) p)
           (or (eq :external ans)
               (format t "~%~%In this Common Lisp, the symbol ~s is ~
                          not among the external symbols of the main ~
                          Lisp package.  Common Lisp requires this.  ~
                          Or maybe we missed one of the CLTL1 ~
                          symbols which were thrown out by CLTL2.  If ~
                          we have, it should be removed from ~
                          *common-lisp-symbols-from-main-lisp-package* ~
                          in axioms-akcl.lisp."
                       sym)))))

(mapcon (function
         (lambda (tl)
           (cond ((or (standard-char-p (car tl))
                      (member (car tl) (cdr tl)))
                  (error "This Common Lisp is unsuitable for Acl2 because ~
                          either (a) one of the three characters #\Page, ~
                          #\Tab, and #\Rubout is said, by standard-char-p, ~
                          to be a standard character or (b) two of the three ~
                          are identical.")))))
        '(#\Page #\Tab #\Rubout))

; If the following code is changed, be sure to change the handling of
; the defconst *directory-separator*.

;; On the toothbrush we take the following for granted... dangerous!!

#|
(let ((namestring (directory-namestring (truename ""))))
  (cond
   ((not (and (stringp namestring)
              (not (equal namestring ""))))
    (error "This Common Lisp is unsuitable for Acl2 because we~%~
            couldn't interpret the current directory as a non-empty~%~
            string, and hence we cannot check that the directory~%~
            separator character is #\:."))
   ((not (eql (char namestring 0)
              #\:))
    (error "This Common Lisp is unsuitable for Acl2 because Acl2~%~
            expects the directory separator character to be #\:, but~%~
            as far as we can tell it is actually ~s."
           (char namestring 0)))))
|#

(dolist (x *features*)
        (cond ((or (equal "AKCL-SET-MV"
                          (symbol-name x))
                   (equal "ACL2-LOGIC-ONLY"
                          (symbol-name x))
                   (equal "ACL2-METERING"
                          (symbol-name x)))
               (format t "~%~%Warning:  This Common Lisp may be ~
                          unsuitable for Acl2 because a symbol with~
                          ~%the name \"ACL2-LOGIC-ONLY\", ~
                          \"AKCL-SET-MV\" or \"ACL2-METERING\" is ~
                          a member of *FEATURES*."))))

(or (typep array-dimension-limit 'fixnum)
    (error "We assume that ARRAY-DIMENSION-LIMIT is a fixnum.  CLTL2 ~
            requires this.  Acl2 will not work in this Common Lisp."))

(or (typep char-code-limit 'fixnum)
    (error "We assume that CHAR-CODE-LIMIT is a fixnum.  Acl2 will ~
            not work in this Common Lisp."))

(defconstant *acl2-package* (find-package "ACL2"))


;                       COMPILING and LOADING

; To compile Acl2, invoke (load "acl2.lisp") and then invoke
; (acl2::compile-acl2).  Having compiled Acl2, to build Acl2, invoke
; (load "acl2.lisp") and then invoke (acl2::load-acl2).  To run Acl2
; verified code, one may either load all of Acl2 as above, or
; alternatively, having complied Acl2, invoke (load "acl2.lisp") and
; invoke (acl2::load-acl2-execution-environment).  The top-level user
; functions for Acl2 are in package "ACL2", so invoke (in-package
; "ACL2") to obviate typing package names.

; The detailed setting of the Acl2 readtable occurs later in this
; file, but we put the declaration of this special before its first
; use.

(defparameter *acl2-readtable*
  (copy-readtable nil)
  "*acl2-readtable* is the readtable we use (a) to prevent the use
#. to cause evaluation during READing (b) and to define our own version
of backquote.")

(defvar *lisp-extension* "lisp")

(defun compile-acl2-toothbrush nil
  (let ((*acl2-files* '("axioms-akcl")))
    (compile-acl2)))

(defun compile-acl2 nil

; Here is a natural place to put compiler options.

; (proclaim '(optimize (safety 0) (space 0) (speed 3)))

; Note on Illegal Instructions:  If Acl2 causes an illegal instruction
; trap it is difficult to figure out what is happening.  Here is a
; brute force way to do it.
; (0) create an events file or book that will recreate the
;     the state up to the event that causes the illegal instruction.
; (1) Copy saved_acl2 to old_saved_acl2
; (2) replace the proclaim above with
;     (proclaim '(optimize (safety 3) (space 0) (speed 0)));
; (3) make full init
;     this will fail because of stack overflows during booting.
; (4) fire up old_saved_acl2
; (5) load the saved events file or book, recreating the state.
; (6) :q from (LP)
; (7) (progn (setq old-world-pair (get 'current-acl2-world 'acl2-world-pair))
;            t)
; (8) (dolist (name *acl2-files*)
;           (or (equal name "defpkgs")
;               (load source)))   ; load the (safety 3) .o files
;     Actually, the last time I did this I loaded each file manually
;     and I did not load ld or interface-raw.  I think they can be
;     loaded here though.
; (9) (progn (f-put-global 'current-acl2-world
;                          (car old-world-pair) *the-live-state*)
;            t)
;(10) (si::use-fast-links nil)
;(11) enter (LP) and execute the offending event.
;(12) once the bug is fixed, be sure to change the proclaim
;     back to (safety 0) (speed 3) before recompiling.

; Note on loading before compiling.  We load each Acl2 source file
; before compiling it to make sure that the functions needed to
; execute macros have been defun-ed before they are called.  Normal
; Common Lisp compilation does not do this.  So we cause all forms to
; be executed before we start the compilation.  This guarantees that
; when macros run, the functions they call have been defined.

; In general, and for the same reason, all Acl2 user checked files
; are also be loaded before they are compiled.

  (let ((*readtable* *acl2-readtable*))
    (dolist (name *acl2-files*)
            (or (equal name "defpkgs")
                (let ((source (make-pathname :name name
                                             :type *lisp-extension*)))
                  (load source)
                  (or (equal name "proof-checker-pkg")
                      (progn
                        (proclaim-file source)
                        (compile-file source)
                        (load name))))))
; ?? (compare-check-sums)
    ))

(defparameter *compiled-file-extension*
  #+akcl "o"
  #+kcl  "o"
  #+lucid "lbin"
  #+allegro "fasl"
  #+mcl "fasl"
  #+symbolics "bin"
  #-(or akcl kcl lucid allegro mcl symbolics)
  (do ((name
        (progn
          (format
           t
           "~%~%We are trying to initialize the parameter that specifies the~%~
            file extension name used by the local compiler.  For example, in~%~
            Allegro Common Lisp, the extension is \"fasl\" while in Lucid~%~
            it is \"lbin\".  Files with such extensions are understood to~%~
            contain compiled code.  We do not recognize this Common Lisp~%~
            from its *features*.  Please type as a STRINGP suitable for READ~%~
            its compiled file extension name:  ")
          (read t))
        (progn
          (format
           t
           "~%~%It has to READ as a STRINGP!  Put quotation marks ~
            around it:  ")
          (read t))))
      ((stringp name) name)))

(defun load-acl2-toothbrush nil
  (let ((*readtable* *acl2-readtable*))
    (load "axioms-akcl")
    (acl2::extend-world1 'acl2::current-acl2-world
                         (acl2::global-set 'acl2::known-package-alist acl2::*initial-known-package-alist*
                                           (acl2::global-set 'acl2::include-book-alist nil
                                                             nil)))
    (in-package "ACL2")
    "ACL2"))

;;; Do not use load-acl2 on the toothbrush.

#|
(defun load-acl2 nil
  (let ((*readtable* *acl2-readtable*))
    (dolist (name *acl2-files*)
            (or (equal name "defpkgs")
                (if (equal name "proof-checker-pkg")
                    (load "proof-checker-pkg.lisp")
                  (load name))))
    (load "defpkgs.lisp")
    (in-package "ACL2")
    "ACL2"))
|#

(defun load-acl2-execution-environment ()
  (let ((*readtable* *acl2-readtable*))
    (load "axioms-akcl")
    (load "defpkgs.lisp")
    (in-package "ACL2")
    "ACL2"))

;                            PROCLAIMING


(defun macroexpand-till (form sym)

; In order to find the THEs that we want to find to do automatic
; proclaiming of the output types of functions, we need to do
; macroexpansion at proclaim-time.  It is possible that a given
; implementation of Common Lisp will macroexpand THE forms further.
; Hence we gently do the macroexpansion we need, one expansion at a
; time, looking for the THE we want to find.

  (loop (cond ((and (consp form) (eq (car form) sym))
               (return form))
              (t (multiple-value-bind
                  (new-form flg)
                  (macroexpand-1 form)
                  (cond ((null flg) (return form))
                        (t (setq form new-form))))))))

(defun get-type-from-dcls (var dcls)
  (cond ((null dcls) t)
        ((and (consp (car dcls))
              (eq 'type (caar dcls))
              (member var (cddr (car dcls))))
         (cadr (car dcls)))
        (t (get-type-from-dcls var (cdr dcls)))))

(defun arg-declarations (formals dcls)
  (cond ((null formals) nil)
        (t (cons (get-type-from-dcls (car formals) dcls)
                 (arg-declarations (cdr formals) dcls)))))

(defun collect-types (l)
  (cond ((null (cdr l)) nil)
        ((stringp (car l))
         (collect-types (cdr l)))
        ((consp (car l))
         (let ((exp (car l)))
           (cond ((and (consp exp) (eq (car exp) 'declare))
                  (append (cdr exp) (collect-types (cdr l))))
                 (t nil))))
        (t nil)))

(defun make-defun-declare-form (form)
  (let* ((formals (caddr form))
         (output (macroexpand-till (car (last form)) 'the))
         (output-type (cond ((and (consp output)
                                  (eq 'the (car output)))
                             (cadr output))
                            (t t))))
    (cond
     ((null (intersection formals lambda-list-keywords))
      `(proclaim (quote (function ,(cadr form)
                                  ,(arg-declarations
                                    formals
                                    (collect-types (cdddr form)))
                                  ,output-type))))
     (t nil))))

(defun make-defconst-declare-form (form)

; We assume that the form has already been evaluated.

  (let* ((output (macroexpand-till (caddr form) 'the))
         (output-type (cond ((and (consp output)
                                  (eq 'the (car output)))
                             (cadr output))
                            (t nil))))
    (cond
     (output-type
      `(proclaim (quote (type ,output-type ,(cadr form)))))
     (t (let ((val (symbol-value (cadr form))))
          (if (integerp val)
              `(proclaim (quote (type (integer ,val ,val) ,(cadr form))))
            nil))))))

(defun proclaim-form (form)
  (cond ((consp form)
         (case (car form)
               ((in-package) (eval form) nil)
               ((defmacro defvar defparameter) nil)
               ((defconst)
                (eval (make-defconst-declare-form form))
                nil)
               ((eval-when)
                (dolist (x (cddr form))
                        (proclaim-form x))
                nil)
               ((progn mutual-recursion)
                (dolist (x (cdr form))
                        (proclaim-form x))
                nil)
               ((defun)
                (eval (make-defun-declare-form form))
                nil)
               (otherwise nil)))
        (t nil)))

(defun proclaim-file (name)

; Proclaims the functions in the file name that are either at the
; top-level, or within a progn, mutual-recursion, or eval-when.
; IMPORTANT:  This function assumes that the defconst forms in the
; given file have already been evaluated.  One way to achieve this
; state of affairs, of course, is to load the file first.

  (with-open-file
   (file name :direction :input)
   (let ((eof (cons nil nil))
         (*package* *package*))
     (loop
      (let ((form (read file nil eof)))
        (cond ((eq eof form) (return nil))
              (t (proclaim-form form)))))
     nil)))

;          Acl2's Implementation of the Backquote Readmacro

(defparameter *comma* (make-symbol "COMMA")
  "*comma* is used by the backquote reader.  When the reader
encounters <,foo>, it returns (list *comma* read:<foo>>).")

(defparameter *comma-atsign* (make-symbol "COMMA-ATSIGN")
  "*comma-atsign* is used by the backquote reader.  When the reader
encounters <,@foo>, it returns (list *comma-atsign* read:<foo>).")

(defparameter *backquote-counter* 0
  "READ cannot handle a comma or comma-atsign unless there is a
pending backquote that will eliminate the *comma* or *comma-atsign*.
In the SPECIAL variable *backquote-counter* we keep track of the number of
backquotes that are currently pending.  It is crucial that this variable
be SPECIAL.")

(defun backquote (x)

  "The two functions BACKQUOTE and BACKQUOTE-LST implement backquote
as described on pp. 349-350 of CLTL1 except that (a) use of vector
notation causes an error and (b) the use of ,. is not permitted."

; It must be emphasized that the Acl2 implementation of backquote is
; only one of many implementations that are consistent with the Common
; Lisp specification of backquote.  That spec requires only that
; backquote produce a form that when evaluated will produce a
; designated object.  We impose the requirement that *acl2-readtable*
; be used both when checking files with Acl2 and when later compiling
; or using those files.  This requirement guarantees that we obtain
; the same behavior of backquote across all Common Lisps.  For
; example, it is an Acl2 theorem, across all Common Lisps that

;   (equal (car '`(,a)) 'cons)

; This theorem is definitely not true about the implementation of
; backquote provided by the implementors of each Common Lisp.  In
; fact, the lefthand side of this theorem represents an informal
; misuse of the backquote notation because one is intended to consider
; the effects of evaluating backquoted forms, not the forms
; themselves.  (In some Common Lisps, the lefthand side might even
; evaluate to a symbol in a nonstandard package.)  Nevertheless,
; because we inflict our definition of backquote on the Acl2 user at
; all times, the above equation is a theorem throughout, so no harm is
; done.  On the other hand, if we used the local implementation of
; backquote of each particular Common Lisp, we would get different
; Acl2 theorems in different Common Lisps, which would be bad.

; Unlike most implementors of backquote, we do no serious
; optimization.  To make matters worse, in Acl2 the backquote
; processing functions run interpreted, not compiled.  We feel this
; inattention to efficiency is justified at the moment because in the
; usage we expect, the only serious costs will be small ones, during
; compilation.

  (cond ((and (vectorp x) (not (stringp x)))
         (error "Acl2 does not handle vectors in backquote."))
        ((atom x) (list 'quote x))
        ((eq (car x) *comma*) (cadr x))
        ((eq (car x) *comma-atsign*) (error "`,@ is an error"))
        (t (backquote-lst x))))

(defun backquote-lst (l)
  (cond ((atom l) (list 'quote l))
        ((eq (car l) *comma*)
         (cadr l))
        ((eq (car l) *comma-atsign*)
         (error ". ,@ is illegal."))
        ((and (consp (car l))
              (eq (caar l) *comma*))
         (list 'cons
               (cadr (car l))
               (backquote-lst (cdr l))))
        ((and (consp (car l))
              (eq (caar l) *comma-atsign*))
         (list 'append (cadr (car l)) (backquote-lst (cdr l))))
        (t
         (list 'cons
               (backquote (car l))
               (backquote-lst (cdr l))))))


;                           ACL2-READTABLE

;  We now make all of the modifications to *acl-readtable*.

(eval-when
 (load eval compile)
 (let ((*readtable* *acl2-readtable*))

; Backquote

   (set-macro-character
    #\`
    #'(lambda (stream char)
        (declare (ignore char))
        (let ((*backquote-counter* (1+ *backquote-counter*)))
          (backquote (read stream t nil t)))))

; Comma

   (set-macro-character
    #\,
    #'(lambda (stream char)
        (declare (ignore char))
        (let ((*backquote-counter* (1- *backquote-counter*)))
          (cond ((< *backquote-counter* 0)
                 (clear-input stream)
                 (error "Illegal comma encountered by READ.")))
          (case (peek-char nil stream t nil t)
                (#\@ (read-char stream t nil t)
                 (list *comma-atsign* (read stream t nil t)))
                (#\. (error ",. not allowed in Acl2 backquote forms."))
                (otherwise (list *comma* (read stream t nil t)))))))

;  Turn off #.

;  Actually, we do not turn it off but make it the all-powerful abort signal.
;  That is, if at any time the user types #. while under the *acl2-readtable*
;  the effect is to abort all the way out to the top-most call of LD.  See LD-FN
;  for some comments.

   (set-dispatch-macro-character
    #\#
    #\.
    #'(lambda (stream char n)

; We used to cause the following hard error in response to #. so pre-empting
; its meaning for abort is no real loss to the Acl2 user.
;       (error "Acl2 does not support #. syntax.")

        (throw 'local-top-level :abort)))

;  Turn off #,

   (set-dispatch-macro-character
    #\#
    #\,
    #'(lambda (stream char n)
        (error "Acl2 does not support #, syntax.")))))


#|
                      Remarks on *acl2-readtable*


Because read-object$ calls the Common Lisp function read, read-object$
is a function of the values of the Common Lisp symbols (a)
*readtable*, (b) *package*, and (c) *features*.  In Acl2, the user can
specify the package to use, via in-package, which sets the global
current-package.  The user of Acl2 has no (legitimate) control over the
readtable, which is set above and discussed below.

As for *features*, we currently permit full use of the *features*
facility of the Common Lisp reader, with this caveat: it is an
official part of the Acl2 story that *features* should have the same
setting throughout all phases of Acl2 usage.  That is, the user must
set up *features* at the beginning, before starting to use Acl2 and
the user must then leave *features* alone (even though the
implementors of Acl2 put :acl2-logic-only onto *features* during
boot-strapping).  One bad consequence of our *features* policy is that
verified files will not in general be verifiable or useable in other
Lisp implementations or installations unless the settings of
*features* relevant to one's usages of the #+ and #- are the same in
the two Lisp implementations.  One simple way to obtain peace of mind
on this topic is for the user not to use #+ or #- at all!  It might be
cleaner for us simply to prohibit the use of the #+ and #- readmacros
all together in Acl2.  This could be done at the cost of replacing the
many uses of #+ and #- in axioms-akcl.lisp, and a few other places, with
some sort of regular macro.

Below is a detailed examination of the default Common Lisp readtable
from the perspective of Acl2.  We want to make sure that when we read,
we do not have side effects (e.g. via #.) of Acl2 but will merely
either (a) cause an error or (b) generate a Lisp object, which we then
will check with bad-lisp-object before handing it to Acl2 functions.

All of the standard Common Lisp characters are either white space or
constituent, with the following exceptions, which are read macros:

  "  quote
  #  number sign
  '  quote
  () parentheses
  ,  comma
  ;  comment
  \  single escape
  `  backquote
  |  multiple escape

With the exception of number sign, backquote and comma, we certainly
want all of these readmacros to have their usual Common Lisp
semantics.

We have defined our own backquote and discussed it above.

We now review the # situation:

  ## and #= are for reading possibly circular stuff
         bad-lisp-object may run forever
  #'  reads as function
         won't hurt anything
  #(  vector
         will be rejected by bad-lisp-object
  #)  signals an error
         enough said
  #*  bit vector
         will be rejected by bad-lisp-object
  #,  load-time evaluation
         we shut it off
  #0-#9 used for infix arguments
         ok
  #:  uninterned symbol
         will be rejected by bad-lisp-object
  #<  signals an error
         enough said
  #\  character object
         will be checked by bad-lisp-object
  #|  start comment, ended by |#
         ok
  #<backspace | tab | newline | linefeed | page | return | space>
      signals an error -- ok
  #+  and #-
      see the discussion of *features* above
  #.  read time evaluation
         we shut it off
  #A, #a  arrays
         will be checked by bad-lisp-object
  #B, #b  binary rational
         ok
  #C, #c complex
         ok (rejected by bad-lisp-object except for rational components)
  #O, #o octal
         ok
  #P, #p pathname
         will be checked by bad-lisp-object
  #R, #r radix-n
         fine
  #S, #s structure
         will be rejected by bad-lisp-object
  #X, #x hex
         ok

Eventually, it will be best to define a read function for Acl2 solely
in terms of Acl2 character reading primitives.  Common Lisp read is
ambiguous.  There is the ambiguity of backquote described above.
There is the abiguity of which tokens get read as numbers.  To make
matters a little more scary, there is nothing that prevents a Common
Lisp implementation from adding, for example, a new # readmacro option
that would provide something as potentially catastrophic as #.  One
obstacle to doing a read within Acl2 this is efficiency.  For example,
Acl2 does not now support unread-char.  And given the requirement that
whatever is specified in a guard must be checkable, it is unclear now
how best to add unread-char since Common Lisp does permit one to
detect whether a stream is in a ``just unread'' state.  Acl2 could
enable one to answer such a question, but at the cost of having to
store the information every time that a character was unread.

|#

