; (proclaim-file "can.lisp")
; (proclaim '(optimize (safety 3) (space 0) (speed 0)))

; The first section of this file defines some symbols that are available in
; full Acl2 but are not available in the miminal version distributed for 
; publication purposes.

(defmacro er (severity ctx string &rest args)
  (cond
   ((eq severity 'hard)
    `(illegal ,ctx ,string (list ,@args)))
   (t `(mv t (list :error ,ctx ,string ,@args) state))))

(defmacro value (x) `(mv nil ,x state))

(defmacro er-let* (alist body)

; This macro introduces the variable er-let-star-use-nowhere-else.
; The user who uses that variable in his forms is likely to be
; disappointed by the fact that we rebind it.

  (declare (xargs :guard (alistp alist)))
  (cond ((null alist) body)
        (t (list 'mv-let
                 (list 'er-let-star-use-nowhere-else
                       (caar alist)
                       'state)
                 (cadar alist)
                 (list 'cond
                       (list 'er-let-star-use-nowhere-else
                             (list 'mv
                                   'er-let-star-use-nowhere-else
                                   (caar alist)
                                   'state))
                       (list t (list 'er-let* (cdr alist) body)))))))

(defconst *nil* ''nil)
(defconst *t*   ''t)

(defun strip-cadrs (x)
  (declare (xargs :guard (all->=-len x 2)))
  (cond ((null x) nil)
        (t (cons (cadar x) (strip-cadrs (cdr x))))))
                                                    
(defun set-difference-eq (l1 l2)
  (cond ((null l1) nil)
        ((member-eq (car l1) l2)
         (set-difference-eq (cdr l1) l2))
        (t (cons (car l1) (set-difference-eq (cdr l1) l2)))))
                                                             
(defun subsetp-eq (x y)
  (cond ((null x) t)
        ((member-eq (car x) y)
         (subsetp-eq (cdr x) y))
        (t nil)))

(defun duplicates (lst)
  (cond ((null lst) nil)
        ((member-eq (car lst) (cdr lst))
         (add-to-set-eq (car lst) (duplicates (cdr lst))))
        (t (duplicates (cdr lst)))))

(defun packn1 (lst)
  (cond ((null lst) nil)
        (t (append (explode-atom (car lst))
                   (packn1 (cdr lst))))))

(defun packn (lst)
  (intern (coerce (packn1 lst) 'string)
          "ACL2"))

; Now we begin with the OBDD work, explicitly.  The next section of this file
; deals with parsing a benchmark file into an untranslated term of the form

; (lambda (input-var1 ... input-vark)
;  (equal (let* ((v1 be1) ...) (list u1 ...))
;         (let* ((vk bek) ...) (list uk ...))

(defun read-input-channel-into-list (chan state)
  (if (f-big-clock-negative-p state)
      (mv 'out-of-time nil state)
      (mv-let (erp val1 state)
              (read-object chan state)
              (if erp
                  (value nil)
                  (let ((state (decrement-big-clock state)))
                    (mv-let (erp rest-vals state)
                            (read-input-channel-into-list chan state)
                            (if erp
                                (mv erp rest-vals state)
                                (value (cons val1 rest-vals)))))))))

(defun read-file-into-list (file-name state)
  (mv-let (chan state)
	  (open-input-channel file-name :object state)
	  (if chan
	      (mv-let (erp lst state)
		      ;; here is where the recursion starts
		      (read-input-channel-into-list chan state)
		      (let ((state (close-input-channel chan state)))
			(if erp
			    (mv erp lst state)
                            (value lst))))
              (er soft 'read-file-into-list "File ~p0 does not seem to exist."
                  file-name))))

(defmacro rf (file-name &optional name)
  (if name
      `(mv-let (erp val state)
	       (read-file-into-list ,file-name state)
	       (if erp
		   (er soft 'rf "File ~p0 does not seem to exist."
		       ,file-name)
                   (assign ,name val)))
      `(read-file-into-list ,file-name state)))

(defun trunc-lst (sym lst)
  (cond ((null lst) nil)
        ((eq sym (car lst)) nil)
        (t (cons (car lst) (trunc-lst sym (cdr lst))))))

(mutual-recursion

 (defun convert-be-expr (expr)
   (cond ((atom expr) expr)
         ((null (cdr expr)) (convert-be-expr (car expr)))
         (t (case (car expr)
                  (not (list 'not (convert-be-expr (cadr expr))))
                  ((and or exor)
                   (cond
                    ((null (cddr expr))

; Apparently (AND p) is a legal be expression.  I assume 
; (AND p) = (OR p) = (EXOR p) = p.

                     (convert-be-expr (cadr expr)))
                    (t
                     (xxxjoin (car expr)
                              (convert-be-expr-lst (cdr expr))))))
                  (t (er hard 'convert-be-expr
                         "Unexpected operand, ~p0."
                         (car expr)))))))

 (defun convert-be-expr-lst (lst)
   (cond ((null lst) nil)
         (t (cons (convert-be-expr (car lst))
                  (convert-be-expr-lst (cdr lst))))))

 )

(defun make-be-bindings (lst)
  (cond ((null lst) nil)
        (t (cons (list (car lst) (convert-be-expr (caddr lst)))
                 (make-be-bindings (cdddr lst))))))

(defun get-be-out-vars (lst)
  (cond ((null lst) nil)
        (t (cons (car lst) (get-be-out-vars (cdddr lst))))))

(mutual-recursion

 (defun compute-used-vars (expr vars)
   (cond ((atom expr)
          (cond ((or (eq expr t) (eq expr nil))
                 vars)
                (t (add-to-set-eq expr vars))))
         (t (compute-used-vars-lst (cdr expr) vars))))

 (defun compute-used-vars-lst (expr-lst vars)
   (cond ((null expr-lst) vars)
         (t (compute-used-vars-lst
             (cdr expr-lst)
             (compute-used-vars (car expr-lst) vars)))))

 )

(defun delete-ignored-bindings (ignored-vars bindings)
  (cond ((null bindings) nil)
        ((member-eq (caar bindings) ignored-vars)
         (delete-ignored-bindings ignored-vars (cdr bindings)))
        (t (cons (car bindings)
                 (delete-ignored-bindings ignored-vars (cdr bindings))))))

(defun parse-be (lst)

; We return (lambda <invars> (let* <bindings> (list <outputs>)))

  (let* ((invars (cadr (member-eq '@invar lst)))
         (subexprs (trunc-lst '@out (cdr (member-eq '@sub lst))))
         (outexprs (trunc-lst '@end (cdr (member-eq '@out lst))))
         (bindings (append (make-be-bindings subexprs)
                           (make-be-bindings outexprs)))
         (outputs (get-be-out-vars outexprs)))
    `(lambda ,invars
       (let* ,(delete-ignored-bindings
               (set-difference-eq
                (strip-cars bindings)
                (union-eq
                 outputs
                 (compute-used-vars (strip-cadrs bindings) nil)))
               bindings)
         (list ,@outputs)))))

(defun set-equal-varsp (lst1 lst2)
  (and (symbolp-listp lst1)
       (no-duplicatesp lst1)
       (symbolp-listp lst2)
       (subsetp-eq lst1 lst2)
       (subsetp-eq lst2 lst1)))

(defun parse-be-file (filename state)

; We return (lambda (<invars>) (equal (let* ...) (let* ...))).

  (er-let* ((lst (rf filename)))
           (let* ((be1 (parse-be lst))
                  (be2 (parse-be (member-eq '@be2 lst)))
                  (input-vars1 (cadr be1))
                  (bindings1 (cadr (caddr be1)))
                  (output-vars1 (cdr (caddr (caddr be1))))
                  (input-vars2 (cadr be2))
                  (bindings2 (cadr (caddr be2)))
                  (output-vars2 (cdr (caddr (caddr be2)))))
             (cond ((not (set-equal-varsp input-vars1 input-vars2))
                    (er soft 'parse-be-file
                        "The input vars of BE1 are ~&0 and those of ~
                         BE2 are ~&1.  But these are supposed to be ~
                         set-equal lists of distinct variable names."
                        input-vars1
                        input-vars2))
                   ((not (no-duplicatesp (strip-cars bindings1)))
                    (er soft 'parse-be-file
                        "The bindings of BE1 contain one or more ~
                         duplications, namely, of ~&0."
                        (duplicates (strip-cars bindings1))))
                   ((not (no-duplicatesp (strip-cars bindings2)))
                    (er soft 'parse-be-file
                        "The bindings of BE2 contain one or more ~
                         duplications, namely, of ~&0."
                        (duplicates (strip-cars bindings2))))
                   ((not (set-equal-varsp output-vars1 output-vars2))
                    (er soft 'parse-be-file
                        "The output vars of BE1 are ~&0 and those of ~
                         BE2 are ~&1.  But these are supposed to be ~
                         set-equal lists of distinct variable names."
                        output-vars1
                        output-vars2))
                   (t
                    (value `(lambda ,input-vars1
                              (equal (let* ,bindings1 (list ,@output-vars1))
                                     (let* ,bindings2 (list ,@output-vars1))))))))))

; Now we move on to the substantive work.

; In this work we represent terms in what we call the cheap syntax.
; Such a "term" is called a "csterm".  Bryant would call it a "node."
; Our goal at the moment is merely to implement a canonicalizer for
; the expressions in the database of bdd benchmarks.  We ultimately
; may convert this to work for more general Acl2 terms, but we are not
; coding it that way now simply so that our own performance measures
; are comparable to published ones.

; Therefore, we are only interested in normalized IF expressions
; composed of variable symbols, the quoted constants t and nil, and
; IF.  We represent variable symbols by natural numbers, the quoted
; constants t and nil by the symbols t and nil, and (IF x y z) by `(k
; ,x ,y . ,z) where k is a natural number that is uniquely associated
; with <x,y,z>.  The association between k and <x,y,z> is arranged via
; a "hash table" discussed below.  The objective is that two
; canonicalized IF expressions are equal (and therefore represent the
; same term) iff their cars are =.

; Macros for accessing canonicalized IFs:

(defmacro unique-id (x) `(the (signed-byte 32) (car ,x)))
(defmacro tst (x) `(the (signed-byte 32) (cadr ,x)))
(defmacro tbr (x) `(caddr ,x))
(defmacro fbr (x) `(cdddr ,x))

; To make cst= very fast, we will represent nil by the if expression
; (0 0 0 . 0) and t will be (1 1 1 . 1).  Thus, two csts are equal
; iff their cars are =.

(defconstant cst-nil '(0 0 0 . 0))
(defconstant cst-t   '(1 1 1 . 1))

; To make this work, we must set the initial mx-id to 1, so the next
; unique id generated is 2.

(defmacro cst= (cst1 cst2)
  `(= (unique-id ,cst1)
      (unique-id ,cst2)))

(defmacro cst-nilp (cst)
  `(= (unique-id ,cst) 0))

(defmacro cst-tp (cst)
  `(= (unique-id ,cst) 1))

; To memo-ize the various merging operations we will hash on the
; opcodes below.  Increasing each by a factor of 3 was intended to
; make it spread out a little more, but it doesn't have much of an
; effect.

(defconstant *id-op* 0)
(defconstant *not-op* (* 3 1))
(defconstant *or-op*  (* 3 2))
(defconstant *and-op* (* 3 3))
(defconstant *xor-op* (* 3 4))
(defconstant *nor-op*  (* 3 5))
(defconstant *nand-op* (* 3 6))
(defconstant *nxor-op* (* 3 7))

; We have two hash arrays, 'if-ht for assigning unique-ids to csts,
; and 'op-ht for memo-izing the merge operators.  Both are hashed on
; triples but the hash code computed is different because we want the
; op-ht code to be commutative wrt the two args but we want the if-ht
; code not to be commutative.

(defmacro if-hash-index (x y z)
  `(the (signed-byte 32)
        (logand (the (signed-byte 32)
                     (+ (the (signed-byte 32)
                             (* 3 (the (signed-byte 32) ,x)))
                        (the (signed-byte 32)
                             (+ (the (signed-byte 32)
                                     (* 2 (unique-id ,y)))
                                (unique-id ,z)))))
                8191)))

(defmacro op-hash-index (op arg1 arg2)
  `(the (signed-byte 32)
        (logand (the (signed-byte 32)
                     (+ (the (signed-byte 32) ,op)
                        (the (signed-byte 32)
                             (+ (unique-id ,arg1)
                                (unique-id ,arg2)))))
                8191)))

; Having found the bucket associated with the hash-index, here is how
; we search it.

(defun if-search-bucket (x y z lst)
  (cond ((null lst) nil)
        ((and (= (the (signed-byte 32) x)
                 (tst (car lst)))
              (cst= y (tbr (car lst)))
              (cst= z (fbr (car lst))))
         (car lst))
        (t (if-search-bucket x y z (cdr lst)))))

(defun op-search-bucket (op arg1 arg2 lst)

; Op-hash-index is commutative wrt arg1 and arg2.  But when we store
; <op, arg1, arg2> in the bucket, we always put the arg with the
; smaller id first.  And when we call this function, arg1 always has
; the smaller id.

  (cond ((null lst) nil)
        ((and (= (the (signed-byte 32) op)
                 (the (signed-byte 32) (cadr (car lst))))
              (cst= arg1 (caddr (car lst)))
              (cst= arg2 (cdddr (car lst))))
         (car (car lst)))
        (t (op-search-bucket op arg1 arg2 (cdr lst)))))

(defun chk-memo (op y z op-ht)

; If <op,y,z> has an entry in op-ht, return T and the entry.
; Otherwise, return NIL and the hash index for <op,y,z> (simply to
; avoid its recomputation).

  (let* ((n (op-hash-index op y z))
         (ans (if (< (unique-id y) (unique-id z))
                  (op-search-bucket op y z (aref1 'op-ht op-ht n))
                  (op-search-bucket op z y (aref1 'op-ht op-ht n)))))
    (cond (ans (mv t ans))
          (t (mv nil n)))))

(defun make-if (n op arg1 arg2 x y z op-ht if-ht mx-id)

; Intuitively, this function makes a cst representing (IF x y z).  But
; we know that this is the answer to the merge op(arg1,arg2) and we
; know that n is the hash index of <op,arg1,arg2>.  We know
; <op,arg1,arg2> is not in the op-ht.  We first look in the if-ht to
; see if (IF x y z) is there.  If so, we return it.  If not, we build
; an appropriate one, assigning the next unique id, which is (1+
; mx-id), and add it to the if-ht.  In any case, before returning, we
; store the returned cst as the answer for op(arg1,arg2) in op-ht.  We
; thus have to return four results: the cst, the two hash arrays, and
; the new mx-id.

  (cond ((cst= y z) (mv y

; The following aset1 was added after my first presentation of this
; work.  Its absence was discovered during a code-walk with Jim
; Bitner.  The times improved slightly on most examples, except mul08
; where we lost a few more seconds.  The times shown in
; ~moore/text/pc-hacking.mss -- the most recent version of the talk on
; this work -- have been updated to show the performance of this
; version of the code.

                        (aset1 'op-ht op-ht n
                               (cons (if (< (unique-id arg1)
                                            (unique-id arg2))
                                         (list* y op arg1 arg2)
                                         (list* y op arg2 arg1))
                                     (aref1 'op-ht op-ht n)))
                        if-ht mx-id))
        (t (let* ((m (if-hash-index x y z))
                  (bucket (aref1 'if-ht if-ht m))
                  (old-if (if-search-bucket x y z bucket)))
             (cond (old-if (mv old-if
                               (aset1 'op-ht op-ht n
                                      (cons (if (< (unique-id arg1)
                                                   (unique-id arg2))
                                                (list* old-if op arg1 arg2)
                                                (list* old-if op arg2 arg1))
                                            (aref1 'op-ht op-ht n)))
                               if-ht
                               mx-id))
                   (t (let ((new-if (list* (1+ (the (signed-byte 32) mx-id))
                                           x y z)))
                        (mv new-if
                            (aset1 'op-ht op-ht n
                                   (cons (if (< (unique-id arg1)
                                                (unique-id arg2))
                                             (list* new-if op arg1 arg2)
                                             (list* new-if op arg2 arg1))
                                         (aref1 'op-ht op-ht n)))
                            (aset1 'if-ht if-ht m
                                   (cons new-if bucket))
                            (car new-if)))))))))

(defun make-not (cst1 op-ht if-ht mx-id)
  (cond ((cst-nilp cst1) (mv cst-t op-ht if-ht mx-id))
        ((cst-tp cst1)   (mv cst-nil op-ht if-ht mx-id))
        (t 
         (mv-let (flg ans)
                 (chk-memo *not-op* cst1 cst-nil op-ht)
                 (cond
                  (flg (mv ans op-ht if-ht mx-id))
                  (t (mv-let (y op-ht if-ht mx-id)
                             (make-not (tbr cst1) op-ht if-ht mx-id)
                             (mv-let (z op-ht if-ht mx-id)
                                     (make-not (fbr cst1) op-ht if-ht mx-id)
                                     (make-if ans *not-op* cst1 cst-nil
                                              (tst cst1)
                                              y
                                              z
                                              op-ht if-ht mx-id)))))))))

(defun make-or (cst1 cst2 op-ht if-ht mx-id)
  (cond
   ((cst-nilp cst1) (mv cst2 op-ht if-ht mx-id))
   ((cst-tp cst1) (mv cst-t op-ht if-ht mx-id))
   ((cst-nilp cst2) (mv cst1 op-ht if-ht mx-id))
   ((cst-tp cst2) (mv cst-t op-ht if-ht mx-id))
   ((cst= cst1 cst2) (mv cst1 op-ht if-ht mx-id))
   (t (mv-let (flg ans)
              (chk-memo *or-op* cst1 cst2 op-ht)
              (cond
               (flg (mv ans op-ht if-ht mx-id))
               (t
                (let ((cst11 (tst cst1))
                      (cst21 (tst cst2)))
                  (cond ((= (the (signed-byte 32) cst11)
                            (the (signed-byte 32) cst21))
                         (mv-let (y op-ht if-ht mx-id)
                                 (make-or (tbr cst1) (tbr cst2)
                                          op-ht if-ht mx-id)
                                 (mv-let (z op-ht if-ht mx-id)
                                         (make-or (fbr cst1) (fbr cst2)
                                                  op-ht if-ht mx-id)
                                         (make-if ans *or-op* cst1 cst2
                                                  cst11 y z
                                                  op-ht if-ht mx-id))))
                        ((< (the (signed-byte 32) cst11)
                            (the (signed-byte 32) cst21))
                         (mv-let (y op-ht if-ht mx-id)
                                 (make-or (tbr cst1) cst2
                                          op-ht if-ht mx-id)
                                 (mv-let (z op-ht if-ht mx-id)
                                         (make-or (fbr cst1) cst2
                                                  op-ht if-ht mx-id)
                                         (make-if ans *or-op* cst1 cst2
                                                  cst11 y z
                                                  op-ht if-ht mx-id))))
                        (t (mv-let (y op-ht if-ht mx-id)
                                   (make-or cst1 (tbr cst2) op-ht if-ht mx-id)
                                   (mv-let (z op-ht if-ht mx-id)
                                           (make-or cst1 (fbr cst2)
                                                    op-ht if-ht mx-id)
                                           (make-if ans *or-op* cst1 cst2
                                                    cst21 y z
                                                    op-ht if-ht mx-id))))))))))))

(defun make-and (cst1 cst2 op-ht if-ht mx-id)
  (cond
   ((cst-nilp cst1) (mv cst-nil op-ht if-ht mx-id))
   ((cst-tp cst1) (mv cst2 op-ht if-ht mx-id))
   ((cst-nilp cst2) (mv cst-nil op-ht if-ht mx-id))
   ((cst-tp cst2) (mv cst1 op-ht if-ht mx-id))
   ((cst= cst1 cst2) (mv cst1 op-ht if-ht mx-id))
   (t (mv-let (flg ans)
              (chk-memo *and-op* cst1 cst2 op-ht)
              (cond
               (flg (mv ans op-ht if-ht mx-id))
               (t
                (let ((cst11 (tst cst1))
                      (cst21 (tst cst2)))
                  (cond ((= (the (signed-byte 32) cst11)
                            (the (signed-byte 32) cst21))
                         (mv-let (y op-ht if-ht mx-id)
                                 (make-and (tbr cst1) (tbr cst2)
                                           op-ht if-ht mx-id)
                                 (mv-let (z op-ht if-ht mx-id)
                                         (make-and (fbr cst1) (fbr cst2)
                                                   op-ht if-ht mx-id)
                                         (make-if ans *and-op* cst1 cst2
                                                  cst11 y z
                                                  op-ht if-ht mx-id))))
                        ((< (the (signed-byte 32) cst11)
                            (the (signed-byte 32) cst21))
                         (mv-let (y op-ht if-ht mx-id)
                                 (make-and (tbr cst1) cst2
                                           op-ht if-ht mx-id)
                                 (mv-let (z op-ht if-ht mx-id)
                                         (make-and (fbr cst1) cst2
                                                   op-ht if-ht mx-id)
                                         (make-if ans *and-op* cst1 cst2
                                                  cst11 y z
                                                  op-ht if-ht mx-id))))
                        (t (mv-let (y op-ht if-ht mx-id)
                                   (make-and cst1 (tbr cst2) op-ht if-ht mx-id)
                                   (mv-let (z op-ht if-ht mx-id)
                                           (make-and cst1 (fbr cst2)
                                                     op-ht if-ht mx-id)
                                           (make-if ans *and-op* cst1 cst2
                                                    cst21 y z
                                                    op-ht if-ht mx-id))))))))))))

(defun make-xor (cst1 cst2 op-ht if-ht mx-id)
  (cond
   ((cst-nilp cst1) (mv cst2 op-ht if-ht mx-id))
   ((cst-tp cst1) (make-not cst2 op-ht if-ht mx-id))
   ((cst-nilp cst2) (mv cst1 op-ht if-ht mx-id))
   ((cst-tp cst2) (make-not cst1 op-ht if-ht mx-id))
   ((cst= cst1 cst2) (mv cst-nil op-ht if-ht mx-id))
   (t (mv-let (flg ans)
              (chk-memo *xor-op* cst1 cst2 op-ht)
              (cond
               (flg (mv ans op-ht if-ht mx-id))
               (t
                (let ((cst11 (tst cst1))
                      (cst21 (tst cst2)))
                  (cond ((= (the (signed-byte 32) cst11)
                            (the (signed-byte 32) cst21))
                         (mv-let (y op-ht if-ht mx-id)
                                 (make-xor (tbr cst1) (tbr cst2)
                                           op-ht if-ht mx-id)
                                 (mv-let (z op-ht if-ht mx-id)
                                         (make-xor (fbr cst1) (fbr cst2)
                                                   op-ht if-ht mx-id)
                                         (make-if ans *xor-op* cst1 cst2
                                                  cst11 y z
                                                  op-ht if-ht mx-id))))
                        ((< (the (signed-byte 32) cst11)
                            (the (signed-byte 32) cst21))
                         (mv-let (y op-ht if-ht mx-id)
                                 (make-xor (tbr cst1) cst2
                                           op-ht if-ht mx-id)
                                 (mv-let (z op-ht if-ht mx-id)
                                         (make-xor (fbr cst1) cst2
                                                   op-ht if-ht mx-id)
                                         (make-if ans *xor-op* cst1 cst2
                                                  cst11 y z
                                                  op-ht if-ht mx-id))))
                        (t (mv-let (y op-ht if-ht mx-id)
                                   (make-xor cst1 (tbr cst2) op-ht if-ht mx-id)
                                   (mv-let (z op-ht if-ht mx-id)
                                           (make-xor cst1 (fbr cst2)
                                                     op-ht if-ht mx-id)
                                           (make-if ans *xor-op* cst1 cst2
                                                    cst21 y z
                                                    op-ht if-ht mx-id))))))))))))

(defun proposition-number (symb a)

; A is an Acl2 one dimensional array.  Each pair in the array (except
; the :header) is of the form (i cbe symb be) where i is a natural
; number, symb is a unique symbol associated with i, be is the
; expression bound to symb in the LET* (or else the symbol INPUT if
; symb is an input variable), and cbe is the canonical form of be (or
; else the canonical form of symb itself if be is INPUT, i.e., (IF
; symb t nil)).  This function returns the i associated with the given
; symb in a.

; Note: We know the :HEADER is of the form (:HEADER :DIMENSION (k)
; :MAXIMUM-LENGTH k :DEFAULT 'X).  Thus, while looking for symb we can
; just take the caddr of each pair in the array and when we get to the
; :HEADER we will be comparing the given symb to '(k), which will
; never be a match and we'll effectively skip the :HEADER without a
; special test.

  (cond ((null a) nil)
        ((eq symb (caddr (car a))) (caar a))
        (t (proposition-number symb (cdr a)))))

; The following two functions are only used to invert cst to
; normalized, canonical terms.

(defun cst (x a)

; a is the array described by proposition-number.  If nil, we make
; up variable names.

  (cond ((cst-nilp x) *nil*)
        ((cst-tp x) *t*)
        ((not (and (integerp (car x))
                   (consp (cdr x))
                   (integerp (cadr x))
                   (consp (cddr x))))
         (er hard 'cst "~p0 is not a cst!" x))
        (t (list 'if
                 (if a
                     (cadr (aref1 'canonicalizer-array a (cadr x)))
                     (packn (list 'v (cadr x))))
                 (cst (caddr x) a)
                 (cst (cdddr x) a)))))

(defun cst-list (x a)
  (cond ((null x) nil)
        (t (cons (cst (car x) a) (cst-list (cdr x) a)))))

; This is for testing purposes only:

(defun recover-bindings (a)
  (cond ((null a) nil)
        ((eq (caar a) :header) (recover-bindings (cdr a)))
        ((eq (cadddr (car a)) 'input) (recover-bindings (cdr a)))
        (t (cons (cddr (car a)) (recover-bindings (cdr a))))))
                              
(defun stuff-bindings (alist expr)

; This function stuffs a bindings-like alist into an untranslated be
; expr and produces an equivalent, translated Acl2 term involving only
; IF, vars and constants.

  (cond ((symbolp expr)
         (cond ((eq expr t) *t*)
               ((eq expr nil) *nil*)
               (t (let ((temp (assoc-eq expr alist)))
                    (cond (temp (stuff-bindings alist (cadr temp)))
                          (t expr))))))
        ((eq (car expr) 'or)
         (list 'if
               (stuff-bindings alist (cadr expr))
               *t*
               (stuff-bindings alist (caddr expr))))
        ((eq (car expr) 'and)
         (list 'if
               (stuff-bindings alist (cadr expr))
               (stuff-bindings alist (caddr expr))
               *nil*))
        ((eq (car expr) 'exor)
         (let ((arg2 (stuff-bindings alist (caddr expr))))
           (list 'if
                 (stuff-bindings alist (cadr expr))
                 (list 'if arg2 *nil* *t*)
                 arg2)))
        ((eq (car expr) 'not)
         (list 'if (stuff-bindings alist (cadr expr)) *nil* *t*))
        (t (er hard 'stuff-bindings "Unrecognized be ~p0." expr))))

(defvar debug-flg nil)
(defvar bad-be)
(defvar bad-a)
(defvar bad-cst)
(defvar bad-be-term)
(defvar bad-cst-term)
(defvar bad-op-ht)
(defvar bad-if-ht)
(defvar bad-mx-id)

; End of testing stuff.

(defun canonicalize (be a op-ht if-ht mx-id)
  (cond ((atom be)
         (cond ((eq be nil) (mv cst-nil op-ht if-ht mx-id))
               ((eq be t) (mv cst-t op-ht if-ht mx-id)) 
               (t (mv (car (aref1 'canonicalizer-array a
                                  (proposition-number be a)))
                      op-ht if-ht mx-id))))
        (t (case (car be)
                 (or
                  (mv-let (y op-ht if-ht mx-id)
                          (canonicalize (cadr be) a op-ht if-ht mx-id)
                          (mv-let (z op-ht if-ht mx-id)
                                  (canonicalize (caddr be) a op-ht if-ht mx-id)
                                  (make-or y z op-ht if-ht mx-id))))
                 (and
                  (mv-let (y op-ht if-ht mx-id)
                          (canonicalize (cadr be) a op-ht if-ht mx-id)
                          (mv-let (z op-ht if-ht mx-id)
                                  (canonicalize (caddr be) a op-ht if-ht mx-id)
                                  (make-and y z op-ht if-ht mx-id))))
                 (exor
                  (mv-let (y op-ht if-ht mx-id)
                          (canonicalize (cadr be) a op-ht if-ht mx-id)
                          (mv-let (z op-ht if-ht mx-id)
                                  (canonicalize (caddr be) a op-ht if-ht mx-id)
                                  (make-xor y z op-ht if-ht mx-id))))
                 (not 
                  (mv-let (y op-ht if-ht mx-id)
                          (canonicalize (cadr be) a op-ht if-ht mx-id)
                          (make-not y op-ht if-ht mx-id)))
                 (t (mv (er hard 'canonicalize "Bad Belgian expression ~p0." be)
                        op-ht if-ht mx-id))))))

(defun canonicalize-list (lst a op-ht if-ht mx-id)
  (cond ((null lst) (mv nil op-ht if-ht mx-id))
        (t (mv-let (x op-ht if-ht mx-id)
                   (canonicalize (car lst) a op-ht if-ht mx-id)
                   (mv-let (y op-ht if-ht mx-id)
                           (canonicalize-list (cdr lst) a op-ht if-ht mx-id)
                           (mv (cons x y) op-ht if-ht mx-id))))))

(defun canonicalize-input-vars (vars i a if-ht mx-id)
  (cond ((null vars) (mv a if-ht mx-id))
        (t (let ((cbe (list* (1+ mx-id) i cst-t cst-nil))
                 (n (if-hash-index i cst-t cst-nil)))
             (canonicalize-input-vars (cdr vars)
                                      (1+ i)
                                      (cons (list i cbe (car vars) 'input)
                                            a)
                                      (aset1 'if-ht if-ht n
                                             (cons cbe (aref1 'if-ht if-ht n)))
                                      (1+ mx-id))))))

(defun canonicalize-bindings (bindings i a op-ht if-ht mx-id)
  (cond ((null bindings) (mv a op-ht if-ht mx-id))
        (t (mv-let (cst op-ht if-ht mx-id)
                   (canonicalize (cadar bindings) a op-ht if-ht mx-id)
                   (canonicalize-bindings (cdr bindings)
                                          (1+ i)
                                          (aset1 'canonicalizer-array
                                                 a
                                                 i
                                                 (cons cst
                                                       (car bindings)))
                                          op-ht if-ht mx-id)))))

(defun canonicalize-let* (nvars x a op-ht if-ht mx-id)

; X is an untranslated let* of the form (LET* ((v1 be1) ...) (LIST u1 ...))
; and nvars is the number of free vars in x.  We canonicalize x.

  (mv-let (a op-ht if-ht mx-id)
          (canonicalize-bindings
           (cadr x)
           nvars
           a op-ht if-ht mx-id)
          (canonicalize-list (cdr (caddr x)) a op-ht if-ht mx-id)))

(defun cst=-lst (lst1 lst2)
  (cond ((null lst1) (null lst2))
        ((cst= (car lst1) (car lst2))
         (cst=-lst (cdr lst1) (cdr lst2)))
        (t nil)))

(defun canonicalize-belgian-expression (x)

; X is an untranslated lambda of the form 
; (LAMBDA vars 
;   (EQUAL (LET* ((v1 be1) ...) (LIST u1 ...))
;          (LET* ((vk bek) ...) (LIST u1 ...))))
; where vars is the list of all free variables occuring in the body.

  (let* ((vars (cadr x))
         (nvars (length vars))
         (dim (+ nvars (max (length (cadr (cadr (caddr x))))
                            (length (cadr (caddr (caddr x))))))))
    (mv-let
     (a if-ht mx-id)
     (canonicalize-input-vars vars
                              0
                              `((:header :dimensions (,dim)
                                         :maximum-length ,dim
                                         :default nil))
                              (compress1 'if-ht
                                         '((:header :dimensions (8192)
                                                    :maximum-length 10000
                                                    :default nil)))
                              1)
     (mv-let (lst1 op-ht if-ht mx-id)
             (canonicalize-let* nvars (cadr (caddr x))
                                (compress1 'canonicalizer-array a)
                                (compress1 'op-ht
                                           '((:header :dimensions (8192)
                                                      :maximum-length 10000
                                                      :default nil)))
                                if-ht mx-id)
             (mv-let (lst2 op-ht if-ht mx-id)
                     (canonicalize-let* nvars (caddr (caddr x))
                                        (compress1 'canonicalizer-array a)
                                        op-ht if-ht mx-id)
                     (mv (cst=-lst lst1 lst2) op-ht if-ht mx-id))))))

(defun bep (x)
  (cond ((symbolp x) t)
        ((and (consp x)
              (true-listp x)
              (equal (length (cdr x))
                     (if (eq (car x) 'not) 1 2)))
         (cond ((eq (car x) 'not)
                (bep (cadr x)))
               ((member-eq (car x) '(and or exor))
                (and (bep (cadr x)) (bep (caddr x))))
               (t nil)))
        (t nil)))

(defun be-input-vars (x)
  (cond ((symbolp x)
         (cond ((or (eq x nil) (eq x t)) nil)
               (t (list x))))
        ((eq (car x) 'not) (be-input-vars (cadr x)))
        (t (union-eq (be-input-vars (cadr x)) (be-input-vars (caddr x))))))

(defun cbe (x y)
  (cond ((not (bep x)) (er hard 'cbe "~p0 is not a bep!" x))
        ((not (bep y)) (er hard 'cbe "~p0 is not a bep!" y))
        (t (canonicalize-belgian-expression
            `(lambda ,(union-eq (be-input-vars x)
                                (be-input-vars y))
               (equal (let* nil (list ,x)) (let* nil (list ,y))))))))


(defun canonicalize-be-file (file state)
  (er-let* ((x (parse-be-file file state)))
           (value (canonicalize-belgian-expression x))))

; In order to make timing tests and debugging easier, we will use

(defvar parse-be-file)

(defvar final-op-ht)
(defvar final-if-ht)
(defvar final-mx-id)

(defun count-ht-collisions (name ht)
  (let (hash-collisions)
    (setq ht (compress1 name ht))
    (setq hash-collisions
          (sort (sloop::sloop for x in (cdr ht)
                              unless (null (cddr x))
                              collect (length (cdr x)))
                '>))
    (list name
          (list :buckets (length (cdr ht)))
          (list :no-of-entries
                (sloop::sloop for x in (cdr ht)
                              sum (length (cdr x))))
          (list :no-of-buckets-of-size-1 (- (length (cdr ht))
                                            (length hash-collisions)))
          (list :bucket-sizes-greater-than-1
                hash-collisions))))

(defun cbf (file)
  (er-let* ((x (parse-be-file file *the-live-state*)))
    (let (y)
      (setq parse-be-file x)
      (time (mv-let (ans op-ht if-ht mx-id)
                    (canonicalize-belgian-expression x)
                    (progn (setq y ans)
                           (setq final-op-ht op-ht)
                           (setq final-if-ht if-ht)
                           (setq final-mx-id mx-id))))
      (cond ((null y)
             (er hard 'cbf "Unequal!"))
            (t (list (list 'mx-id final-mx-id)
                     (count-ht-collisions 'op-ht final-op-ht)
                     (count-ht-collisions 'if-ht final-if-ht)))))))

; The files listed here are all the .be files on cath/ and on ex/.
; There are still other .be files on hachtel/ for which we have
; published comparisons, but I am limiting my attention to these
; for the moment.

(defvar *benchmark-files*
  '("cath/add1.be"
    "cath/add2.be"
    "cath/add3.be"
    "cath/add4.be"
    "cath/addsub.be"
;   "cath/alu.be"     ; has a DCS, so ignored now
;   "ex/ex2.be"       ; has a DCS, so ignored now
    "ex/mul03.be"
    "ex/mul04.be"
    "ex/mul05.be"
    "ex/mul06.be"
    "ex/mul07.be"
    "ex/mul08.be"
    "ex/rip02.be"
    "ex/rip04.be"
    "ex/rip06.be"
    "ex/rip08.be"
    "ex/transp.be"
    "ex/ztwaalf1.be"
    "ex/ztwaalf2.be"))

(defun allocations nil
  (si::allocate 'cons 10000 t)
  (si::allocate 'fixnum 1000 t))

(defun cbfs ()
  (allocations)
  (sloop::sloop for x in *benchmark-files*
                do (si::gbc t)  ; Level 3 garbage collection before starting
                   (print x)
                   (print (cbf x))
                   (terpri)))

#|
ACL2>(cbfs)
[GC for 0 RELOCATABLE-BLOCKS pages..(T=153).GC finished]
[GC for 0 RELOCATABLE-BLOCKS pages..(T=192).GC finished]
[GC for 0 CONTIGUOUS-BLOCKS pages..(T=221).GC finished]

"cath/add1.be" 
real time : 0.333 secs
run time  : 0.150 secs

((MX-ID 432)
 (OP-HT (:BUCKETS 382) (:NO-OF-ENTRIES 640)
        (:NO-OF-BUCKETS-OF-SIZE-1 222)
        (:BUCKET-SIZES-GREATER-THAN-1
            (6 5 5 5 5 5 5 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2)))
 (IF-HT (:BUCKETS 334) (:NO-OF-ENTRIES 431)
        (:NO-OF-BUCKETS-OF-SIZE-1 250)
        (:BUCKET-SIZES-GREATER-THAN-1
            (4 3 3 3 3 3 3 3 3 3 3 3 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2)))) 
[GC for 0 CONTIGUOUS-BLOCKS pages..(T=229).GC finished]

"cath/add2.be" 
real time : 0.633 secs
run time  : 0.617 secs

((MX-ID 1876)
 (OP-HT (:BUCKETS 1647) (:NO-OF-ENTRIES 3076)
        (:NO-OF-BUCKETS-OF-SIZE-1 955)
        (:BUCKET-SIZES-GREATER-THAN-1
            (10 10 9 9 9 9 8 8 8 8 8 8 7 7 7 7 7 7 7 7 7 7 7 7 7 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2)))
 (IF-HT (:BUCKETS 1410) (:NO-OF-ENTRIES 1875)
        (:NO-OF-BUCKETS-OF-SIZE-1 1116)
        (:BUCKET-SIZES-GREATER-THAN-1
            (8 7 7 6 6 5 5 5 5 5 5 5 5 5 5 5 5 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2)))) 
[GC for 0 CONTIGUOUS-BLOCKS pages..(T=229).GC finished]

"cath/add3.be" 
real time : 1.783 secs
run time  : 1.700 secs

((MX-ID 4806)
 (OP-HT (:BUCKETS 4355) (:NO-OF-ENTRIES 8356)
        (:NO-OF-BUCKETS-OF-SIZE-1 2026)
        (:BUCKET-SIZES-GREATER-THAN-1
            (9 9 8 8 8 8 8 7 7 7 7 7 7 7 7 7 7 7 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2)))
 (IF-HT (:BUCKETS 3047) (:NO-OF-ENTRIES 4805)
        (:NO-OF-BUCKETS-OF-SIZE-1 1958)
        (:BUCKET-SIZES-GREATER-THAN-1
            (8 6 6 6 6 6 6 6 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2)))) 
[GC for 0 CONTIGUOUS-BLOCKS pages..(T=238).GC finished]

"cath/add4.be" 
real time : 3.933 secs
run time  : 3.917 secs

((MX-ID 10820)
 (OP-HT (:BUCKETS 6759) (:NO-OF-ENTRIES 18384)
        (:NO-OF-BUCKETS-OF-SIZE-1 2211)
        (:BUCKET-SIZES-GREATER-THAN-1
            (17 16 15 15 14 14 13 13 12 12 11 11 11 11 11 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2)))
 (IF-HT (:BUCKETS 5299) (:NO-OF-ENTRIES 10819)
        (:NO-OF-BUCKETS-OF-SIZE-1 2451)
        (:BUCKET-SIZES-GREATER-THAN-1
            (19 16 12 12 10 9 9 9 9 9 9 9 9 8 8 8 8 8 8 8 8 8 8 8 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2)))) 
[GC for 0 CONTIGUOUS-BLOCKS pages..(T=255).GC finished]

"cath/addsub.be" 
real time : 1.483 secs
run time  : 1.467 secs

((MX-ID 4254)
 (OP-HT (:BUCKETS 4308) (:NO-OF-ENTRIES 6716)
        (:NO-OF-BUCKETS-OF-SIZE-1 2593)
        (:BUCKET-SIZES-GREATER-THAN-1
            (6 5 5 5 5 5 5 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2)))
 (IF-HT (:BUCKETS 2493) (:NO-OF-ENTRIES 4253)
        (:NO-OF-BUCKETS-OF-SIZE-1 1382)
        (:BUCKET-SIZES-GREATER-THAN-1
            (5 5 5 5 5 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2)))) 
[GC for 0 CONTIGUOUS-BLOCKS pages..(T=243).GC finished]

"ex/mul03.be" 
real time : 0.050 secs
run time  : 0.050 secs

((MX-ID 114)
 (OP-HT (:BUCKETS 93) (:NO-OF-ENTRIES 157)
        (:NO-OF-BUCKETS-OF-SIZE-1 53)
        (:BUCKET-SIZES-GREATER-THAN-1
            (4 4 4 4 4 3 3 3 3 3 3 3 3 3 3 3 3 3 3 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2)))
 (IF-HT (:BUCKETS 94) (:NO-OF-ENTRIES 113)
        (:NO-OF-BUCKETS-OF-SIZE-1 77)
        (:BUCKET-SIZES-GREATER-THAN-1
            (3 3 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2)))) 
[GC for 0 CONTIGUOUS-BLOCKS pages..(T=227).GC finished]

"ex/mul04.be" 
real time : 0.233 secs
run time  : 0.233 secs

((MX-ID 723)
 (OP-HT (:BUCKETS 675) (:NO-OF-ENTRIES 1241)
        (:NO-OF-BUCKETS-OF-SIZE-1 370)
        (:BUCKET-SIZES-GREATER-THAN-1
            (7 7 7 7 7 7 7 6 6 6 6 6 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2)))
 (IF-HT (:BUCKETS 590) (:NO-OF-ENTRIES 722)
        (:NO-OF-BUCKETS-OF-SIZE-1 474)
        (:BUCKET-SIZES-GREATER-THAN-1
            (3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2)))) 
[GC for 0 CONTIGUOUS-BLOCKS pages..(T=227).GC finished]

"ex/mul05.be" 
real time : 1.067 secs
run time  : 1.050 secs

((MX-ID 3173)
 (OP-HT (:BUCKETS 3042) (:NO-OF-ENTRIES 5533)
        (:NO-OF-BUCKETS-OF-SIZE-1 1713)
        (:BUCKET-SIZES-GREATER-THAN-1
            (10 10 10 9 9 9 9 9 9 9 8 8 8 8 8 8 8 8 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2)))
 (IF-HT (:BUCKETS 2496) (:NO-OF-ENTRIES 3172)
        (:NO-OF-BUCKETS-OF-SIZE-1 1962)
        (:BUCKET-SIZES-GREATER-THAN-1
            (5 5 5 5 5 5 5 5 5 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2)))) 
[GC for 0 CONTIGUOUS-BLOCKS pages..(T=235).GC finished]

"ex/mul06.be" 
real time : 4.817 secs
run time  : 4.783 secs

((MX-ID 12595)
 (OP-HT (:BUCKETS 7374) (:NO-OF-ENTRIES 22967)
        (:NO-OF-BUCKETS-OF-SIZE-1 1630)
        (:BUCKET-SIZES-GREATER-THAN-1
            (15 14 14 14 13 13 13 13 13 13 13 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2)))
 (IF-HT (:BUCKETS 6379) (:NO-OF-ENTRIES 12594)
        (:NO-OF-BUCKETS-OF-SIZE-1 2736)
        (:BUCKET-SIZES-GREATER-THAN-1
            (9 9 9 8 8 8 8 8 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2)))) 
[GC for 0 CONTIGUOUS-BLOCKS pages..(T=262).GC finished]

"ex/mul07.be" [GC for 1000 FIXNUM pages..(T=323).GC finished]

real time : 26.233 secs
run time  : 25.783 secs

((MX-ID 43695)
 (OP-HT (:BUCKETS 8192) (:NO-OF-ENTRIES 81025)
        (:NO-OF-BUCKETS-OF-SIZE-1 6)
        (:BUCKET-SIZES-GREATER-THAN-1
            (26 25 25 25 24 24 24 24 23 23 23 23 23 23 22 22 22 22 22
             22 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 20 20
             20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20
             20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20
             20 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19
             19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2)))
 (IF-HT (:BUCKETS 8153) (:NO-OF-ENTRIES 43694)
        (:NO-OF-BUCKETS-OF-SIZE-1 212)
        (:BUCKET-SIZES-GREATER-THAN-1
            (18 15 15 15 14 14 14 14 14 14 14 13 13 13 13 13 13 13 13
             13 13 13 13 13 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7
             7 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6 6
             6 6 6 6 6 6 6 6 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5
             5 5 5 5 5 5 5 5 5 5 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2)))) 
[GC for 0 CONTIGUOUS-BLOCKS pages..(T=322).GC finished]

"ex/mul08.be" [GC for 1000 FIXNUM pages..(T=374).GC finished]
[GC for 1000 FIXNUM pages..(T=404).GC finished]
[GC for 1000 FIXNUM pages..(T=441).GC finished]
[GC for 1000 FIXNUM pages..(T=467).GC finished]
[GC for 1000 FIXNUM pages..(T=488).GC finished]
[GC for 1000 FIXNUM pages..(T=509).GC finished]
[GC for 1000 FIXNUM pages..(T=531).GC finished]
[GC for 1000 FIXNUM pages..(T=551).GC finished]

real time : 154.900 secs
run time  : 151.183 secs

((MX-ID 144767)
 (OP-HT (:BUCKETS 8192) (:NO-OF-ENTRIES 272000)
        (:NO-OF-BUCKETS-OF-SIZE-1 0)
        (:BUCKET-SIZES-GREATER-THAN-1
            (60 58 57 55 55 54 53 53 53 53 52 52 52 52 52 51 51 51 51
             50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50
             50 50 50 50 50 50 49 49 49 49 49 49 49 49 49 49 49 49 49
             49 49 49 49 49 49 49 49 49 49 49 49 49 49 49 49 49 48 48
             48 48 48 48 48 48 48 48 48 48 48 48 48 48 48 48 48 48 48
             48 48 48 48 48 48 48 48 48 48 48 48 48 48 48 48 48 48 48
             48 48 48 48 48 48 47 47 47 47 47 47 47 47 47 47 47 47 47
             47 47 47 47 47 47 47 47 47 47 47 47 47 47 47 47 47 47 47
             47 47 47 47 47 47 47 47 47 47 47 47 47 47 47 47 47 47 47
             47 47 47 47 47 47 47 47 47 47 47 47 47 47 47 47 46 46 46
             46 46 46 46 46 46 46 46 46 46 46 46 46 46 46 46 46 46 46
             46 46 46 46 46 46 46 46 46 46 46 46 46 46 46 46 46 46 46
             46 46 46 46 46 46 46 46 46 46 46 46 46 46 46 46 46 46 46
             46 46 46 46 46 45 45 45 45 45 45 45 45 45 45 45 45 45 45
             45 45 45 45 45 45 45 45 45 45 45 45 45 45 45 45 45 45 45
             45 45 45 45 45 45 45 45 45 45 45 45 45 45 45 45 45 45 45
             45 45 45 45 45 45 45 45 45 45 45 45 45 45 45 45 45 45 45
             45 45 45 45 45 45 45 45 45 45 45 45 45 45 45 45 45 45 45
             45 45 45 45 45 45 45 45 45 45 44 44 44 44 44 44 44 44 44
             44 44 44 44 44 44 44 44 44 44 44 44 44 44 44 44 44 44 44
             44 44 44 44 44 44 44 44 44 44 44 44 44 44 44 44 44 44 44
             44 44 44 44 44 44 44 44 44 44 44 44 44 44 44 44 44 44 44
             44 44 44 44 44 44 44 44 44 44 44 44 44 44 44 44 44 44 44
             44 44 44 44 44 44 44 44 44 44 44 44 44 44 44 44 44 44 44
             44 44 44 44 44 44 44 44 44 44 44 44 44 44 44 44 44 44 44
             44 44 44 44 44 44 44 44 44 44 44 44 44 44 44 44 44 44 44
             44 44 44 44 43 43 43 43 43 43 43 43 43 43 43 43 43 43 43
             43 43 43 43 43 43 43 43 43 43 43 43 43 43 43 43 43 43 43
             43 43 43 43 43 43 43 43 43 43 43 43 43 43 43 43 43 43 43
             43 43 43 43 43 43 43 43 43 43 43 43 43 43 43 43 43 43 43
             43 43 43 43 43 43 43 43 43 43 43 43 43 43 43 43 43 43 43
             43 43 43 43 43 43 43 43 43 43 43 43 43 43 43 43 43 43 43
             43 43 43 43 43 43 43 43 43 43 43 43 43 43 43 43 43 43 43
             43 43 43 43 43 43 43 43 43 43 43 43 43 43 43 43 43 43 43
             43 43 43 43 43 43 43 43 43 43 43 43 43 43 43 43 43 43 42
             42 42 42 42 42 42 42 42 42 42 42 42 42 42 42 42 42 42 42
             42 42 42 42 42 42 42 42 42 42 42 42 42 42 42 42 42 42 42
             42 42 42 42 42 42 42 42 42 42 42 42 42 42 42 42 42 42 42
             42 42 42 42 42 42 42 42 42 42 42 42 42 42 42 42 42 42 42
             42 42 42 42 42 42 42 42 42 42 42 42 42 42 42 42 42 42 42
             42 42 42 42 42 42 42 42 42 42 42 42 42 42 42 42 42 42 42
             42 42 42 42 42 42 42 42 42 42 42 42 42 42 42 42 42 42 42
             42 42 42 42 42 42 42 42 42 42 42 42 42 42 42 42 42 42 42
             42 42 42 42 42 42 42 42 42 42 42 42 42 42 42 42 42 42 42
             42 42 42 42 42 42 42 42 42 42 42 42 42 42 42 42 42 42 42
             42 42 42 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41
             41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41
             41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41
             41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41
             41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41
             41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41
             41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41
             41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41
             41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41
             41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41
             41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41
             41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41
             41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 41 40
             40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40
             40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40
             40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40
             40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40
             40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40
             40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40
             40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40
             40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40
             40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40
             40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40
             40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40
             40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40
             40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40
             40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40
             40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40 40
             40 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39
             39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39
             39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39
             39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39
             39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39
             39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39
             39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39
             39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39
             39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39
             39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39
             39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39
             39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39
             39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39
             39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39
             39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39
             39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39
             39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39
             39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39 39
             39 39 39 39 39 39 39 39 39 39 38 38 38 38 38 38 38 38 38
             38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38
             38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38
             38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38
             38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38
             38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38
             38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38
             38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38
             38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38
             38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38
             38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38
             38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38
             38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38
             38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38
             38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38
             38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38
             38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38
             38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38
             38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38 38
             38 38 38 38 38 37 37 37 37 37 37 37 37 37 37 37 37 37 37
             37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37
             37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37
             37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37
             37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37
             37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37
             37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37
             37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37
             37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37
             37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37
             37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37
             37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37
             37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37
             37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37
             37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37
             37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37
             37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37
             37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37
             37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37
             37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37
             37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37
             37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 37 36 36
             36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36
             36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36
             36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36
             36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36
             36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36
             36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36
             36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36
             36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36
             36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36
             36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36
             36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36
             36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36
             36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36
             36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36
             36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36
             36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36
             36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36
             36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36
             36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36
             36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36
             36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36
             36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36
             36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36 36
             36 36 36 36 36 36 36 35 35 35 35 35 35 35 35 35 35 35 35
             35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35
             35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35
             35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35
             35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35
             35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35
             35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35
             35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35
             35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35
             35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35
             35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35
             35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35
             35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35
             35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35
             35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35
             35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35
             35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35
             35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35
             35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35
             35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35
             35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35
             35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35 35
             35 35 35 35 35 35 35 35 35 35 35 34 34 34 34 34 34 34 34
             34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34
             34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34
             34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34
             34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34
             34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34
             34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34
             34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34
             34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34
             34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34
             34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34
             34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34
             34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34
             34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34
             34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34
             34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34
             34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34
             34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34
             34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34
             34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34
             34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34
             34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34
             34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34
             34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34
             34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34
             34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34 34
             34 34 34 34 34 34 34 34 34 34 33 33 33 33 33 33 33 33 33
             33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33
             33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33
             33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33
             33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33
             33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33
             33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33
             33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33
             33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33
             33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33
             33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33
             33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33
             33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33
             33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33
             33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33
             33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33
             33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33
             33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33
             33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33
             33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33
             33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33
             33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33
             33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33
             33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33
             33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33
             33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33
             33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 33 32 32 32
             32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32
             32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32
             32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32
             32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32
             32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32
             32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32
             32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32
             32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32
             32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32
             32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32
             32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32
             32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32
             32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32
             32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32
             32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32
             32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32
             32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32
             32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32
             32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32
             32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32
             32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32
             32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32
             32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32
             32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32 32
             32 32 32 32 32 32 32 32 32 32 32 31 31 31 31 31 31 31 31
             31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31
             31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31
             31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31
             31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31
             31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31
             31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31
             31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31
             31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31
             31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31
             31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31
             31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31
             31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31
             31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31
             31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31
             31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31
             31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31
             31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31
             31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31
             31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31
             31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31
             31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31
             31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31
             31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31
             31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31
             31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31 31
             31 31 31 31 31 31 31 31 31 31 31 31 31 30 30 30 30 30 30
             30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30
             30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30
             30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30
             30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30
             30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30
             30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30
             30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30
             30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30
             30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30
             30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30
             30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30
             30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30
             30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30
             30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30
             30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30
             30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30
             30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30
             30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30
             30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30
             30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30
             30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30
             30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30
             30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30
             30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30
             30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 29
             29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29
             29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29
             29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29
             29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29
             29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29
             29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29
             29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29
             29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29
             29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29
             29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29
             29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29
             29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29
             29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29
             29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29
             29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29
             29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29
             29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29
             29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29
             29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29
             29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29
             29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29
             29 29 29 29 29 29 29 29 29 29 29 29 29 29 29 28 28 28 28
             28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28
             28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28
             28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28
             28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28
             28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28
             28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28
             28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28
             28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28
             28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28
             28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28
             28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28
             28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28
             28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28
             28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28
             28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28
             28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28
             28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28
             28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28
             28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28
             28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28
             28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28
             28 28 28 28 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27
             27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27
             27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27
             27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27
             27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27
             27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27
             27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27
             27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27
             27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27
             27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27
             27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27
             27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27
             27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27
             27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27
             27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27
             27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27
             27 27 27 27 27 27 27 27 27 26 26 26 26 26 26 26 26 26 26
             26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26
             26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26
             26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26
             26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26
             26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26
             26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26
             26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26
             26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26
             26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26
             26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26
             26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26
             26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26
             26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26
             26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26
             26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26
             26 26 26 26 26 26 26 26 26 26 26 25 25 25 25 25 25 25 25
             25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25
             25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25
             25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25
             25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25
             25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25
             25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25
             25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25
             25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25
             25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25
             25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25
             25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25
             25 25 25 25 25 25 25 25 24 24 24 24 24 24 24 24 24 24 24
             24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24
             24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24
             24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24
             24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24
             24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24
             24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24
             24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24
             24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24
             24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24
             24 24 24 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23
             23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23
             23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23
             23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23
             23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23
             23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23
             23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23
             23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23
             23 23 23 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22
             22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22
             22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22
             22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22
             22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22
             22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22
             22 22 22 22 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21
             21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21
             21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21
             21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21
             21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 20 20 20
             20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20
             20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20
             20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20
             20 20 20 20 20 19 19 19 19 19 19 19 19 19 19 19 19 19 19
             19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19
             19 19 19 19 19 19 19 19 19 19 19 19 19 18 18 18 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18
             18 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             16 16 16 16 16 16 15 15 15 15 15 15 15 14 14 14 14 14 13
             12 12 12)))
 (IF-HT (:BUCKETS 8192) (:NO-OF-ENTRIES 144766)
        (:NO-OF-BUCKETS-OF-SIZE-1 0)
        (:BUCKET-SIZES-GREATER-THAN-1
            (33 33 33 33 33 33 33 32 32 32 32 32 32 32 31 31 31 31 31
             31 31 31 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30 30
             30 30 30 30 29 29 29 29 29 29 29 29 29 29 29 29 29 29 29
             29 29 29 29 29 29 29 29 29 28 28 28 28 28 28 28 28 28 28
             28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28
             28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28 28
             28 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27
             27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27
             27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 27 26 26 26
             26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26
             26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26
             26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26
             26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26
             26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26
             26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26
             26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26 26
             26 26 26 26 26 26 26 25 25 25 25 25 25 25 25 25 25 25 25
             25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25
             25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25
             25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25
             25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25
             25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25
             25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25
             25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25
             25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25 25
             25 25 25 25 25 24 24 24 24 24 24 24 24 24 24 24 24 24 24
             24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24
             24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24
             24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24
             24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24
             24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24
             24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24
             24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24
             24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24
             24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24
             24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24
             24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24
             24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24
             24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24 24
             24 24 24 24 24 24 24 24 24 24 24 24 24 24 23 23 23 23 23
             23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23
             23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23
             23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23
             23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23
             23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23
             23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23
             23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23
             23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23
             23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23
             23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23
             23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23
             23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23
             23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23
             23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23
             23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23
             23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23
             23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23 23
             23 23 23 23 23 23 23 23 23 23 22 22 22 22 22 22 22 22 22
             22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22
             22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22
             22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22
             22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22
             22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22
             22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22
             22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22
             22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22
             22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22
             22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22
             22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22
             22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22
             22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22
             22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22
             22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22
             22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22
             22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22
             22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22
             22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22 22
             22 22 22 22 22 22 22 22 22 22 22 22 22 22 21 21 21 21 21
             21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21
             21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21
             21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21
             21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21
             21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21
             21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21
             21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21
             21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21
             21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21
             21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21
             21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21
             21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21
             21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21
             21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21
             21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21
             21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21
             21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21
             21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21
             21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21
             21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21
             21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21
             21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21
             21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21
             21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21
             21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21
             21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21
             21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21
             21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21 21
             21 21 21 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20
             20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20
             20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20
             20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20
             20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20
             20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20
             20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20
             20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20
             20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20
             20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20
             20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20
             20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20
             20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20
             20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20
             20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20
             20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20
             20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20
             20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20
             20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20
             20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20
             20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20
             20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20
             20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20
             20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20
             20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20
             20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20
             20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20
             20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20
             20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20
             20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20
             20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20
             20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20
             20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20 20
             20 20 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19
             19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19
             19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19
             19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19
             19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19
             19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19
             19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19
             19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19
             19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19
             19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19
             19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19
             19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19
             19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19
             19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19
             19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19
             19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19
             19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19
             19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19
             19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19
             19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19
             19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19
             19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19
             19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19
             19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19
             19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19
             19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19
             19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19
             19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19
             19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19
             19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19
             19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19
             19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19
             19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19
             19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19
             19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19
             19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19 19
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18 18
             18 18 18 18 18 18 18 18 18 18 18 18 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17 17
             17 17 17 17 17 17 17 17 17 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
             16 16 16 16 16 16 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 15
             15 15 15 15 15 15 15 15 15 15 15 15 15 15 15 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14 14
             14 14 14 14 14 14 14 14 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13 13
             13 13 13 13 13 13 13 13 13 13 13 13 13 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12
             12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 12 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11 11
             11 11 11 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10
             10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9 9
             9 9 9 9 9 9 9 9 9 9 9 9 9 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
             8 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 7 6 6 6 6 6 6 6 6
             6 6 6 6 5 5 4)))) 
[GC for 0 CONTIGUOUS-BLOCKS pages..(T=492).GC finished]

"ex/rip02.be" 
real time : 0.033 secs
run time  : 0.017 secs

((MX-ID 39)
 (OP-HT (:BUCKETS 36) (:NO-OF-ENTRIES 46) (:NO-OF-BUCKETS-OF-SIZE-1 27)
        (:BUCKET-SIZES-GREATER-THAN-1 (3 2 2 2 2 2 2 2 2)))
 (IF-HT (:BUCKETS 28) (:NO-OF-ENTRIES 38) (:NO-OF-BUCKETS-OF-SIZE-1 20)
        (:BUCKET-SIZES-GREATER-THAN-1 (3 3 2 2 2 2 2 2)))) 
[GC for 0 CONTIGUOUS-BLOCKS pages..(T=258).GC finished]

"ex/rip04.be" 
real time : 0.067 secs
run time  : 0.067 secs

((MX-ID 196)
 (OP-HT (:BUCKETS 166) (:NO-OF-ENTRIES 239)
        (:NO-OF-BUCKETS-OF-SIZE-1 122)
        (:BUCKET-SIZES-GREATER-THAN-1
            (5 5 5 4 4 4 3 3 3 3 3 3 3 3 3 3 3 3 3 3 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2)))
 (IF-HT (:BUCKETS 170) (:NO-OF-ENTRIES 195)
        (:NO-OF-BUCKETS-OF-SIZE-1 148)
        (:BUCKET-SIZES-GREATER-THAN-1
            (3 3 3 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2)))) 
[GC for 0 CONTIGUOUS-BLOCKS pages..(T=226).GC finished]

"ex/rip06.be" 
real time : 0.133 secs
run time  : 0.133 secs

((MX-ID 485)
 (OP-HT (:BUCKETS 384) (:NO-OF-ENTRIES 588)
        (:NO-OF-BUCKETS-OF-SIZE-1 266)
        (:BUCKET-SIZES-GREATER-THAN-1
            (7 6 6 5 5 5 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2)))
 (IF-HT (:BUCKETS 394) (:NO-OF-ENTRIES 484)
        (:NO-OF-BUCKETS-OF-SIZE-1 312)
        (:BUCKET-SIZES-GREATER-THAN-1
            (3 3 3 3 3 3 3 3 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2)))) 
[GC for 0 CONTIGUOUS-BLOCKS pages..(T=225).GC finished]

"ex/rip08.be" 
real time : 0.283 secs
run time  : 0.250 secs

((MX-ID 906)
 (OP-HT (:BUCKETS 702) (:NO-OF-ENTRIES 1093)
        (:NO-OF-BUCKETS-OF-SIZE-1 476)
        (:BUCKET-SIZES-GREATER-THAN-1
            (7 6 6 6 5 5 5 5 5 5 5 5 5 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4
             4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 4 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 3 3 3 3 3 3 3 3 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2)))
 (IF-HT (:BUCKETS 718) (:NO-OF-ENTRIES 905)
        (:NO-OF-BUCKETS-OF-SIZE-1 546)
        (:BUCKET-SIZES-GREATER-THAN-1
            (4 3 3 3 3 3 3 3 3 3 3 3 3 3 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2)))) 
[GC for 0 CONTIGUOUS-BLOCKS pages..(T=228).GC finished]

"ex/transp.be" 
real time : 0.017 secs
run time  : 0.017 secs

((MX-ID 19)
 (OP-HT (:BUCKETS 19) (:NO-OF-ENTRIES 20) (:NO-OF-BUCKETS-OF-SIZE-1 18)
        (:BUCKET-SIZES-GREATER-THAN-1 (2)))
 (IF-HT (:BUCKETS 16) (:NO-OF-ENTRIES 18) (:NO-OF-BUCKETS-OF-SIZE-1 14)
        (:BUCKET-SIZES-GREATER-THAN-1 (2 2)))) 
[GC for 0 CONTIGUOUS-BLOCKS pages..(T=224).GC finished]

"ex/ztwaalf1.be" 
real time : 0.083 secs
run time  : 0.083 secs

((MX-ID 309)
 (OP-HT (:BUCKETS 258) (:NO-OF-ENTRIES 405)
        (:NO-OF-BUCKETS-OF-SIZE-1 157)
        (:BUCKET-SIZES-GREATER-THAN-1
            (5 5 4 4 4 4 4 4 4 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             3 3 3 3 3 3 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2)))
 (IF-HT (:BUCKETS 258) (:NO-OF-ENTRIES 308)
        (:NO-OF-BUCKETS-OF-SIZE-1 214)
        (:BUCKET-SIZES-GREATER-THAN-1
            (3 3 3 3 3 3 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2)))) 
[GC for 0 CONTIGUOUS-BLOCKS pages..(T=223).GC finished]

"ex/ztwaalf2.be" 
real time : 0.100 secs
run time  : 0.083 secs

((MX-ID 309)
 (OP-HT (:BUCKETS 267) (:NO-OF-ENTRIES 405)
        (:NO-OF-BUCKETS-OF-SIZE-1 166)
        (:BUCKET-SIZES-GREATER-THAN-1
            (4 4 4 4 4 4 4 4 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3 3
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2 2 2 2)))
 (IF-HT (:BUCKETS 264) (:NO-OF-ENTRIES 308)
        (:NO-OF-BUCKETS-OF-SIZE-1 224)
        (:BUCKET-SIZES-GREATER-THAN-1
            (3 3 3 3 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2 2
             2 2 2 2 2 2 2 2 2 2 2)))) 
|#